"""
Model classes used by the Workspaces system.
"""
import abc
import enum
import json
from pathlib import Path
from typing import Dict, List, Tuple

import sqlalchemy as sa
from pycapo import CapoConfig
from sqlalchemy import create_engine
from sqlalchemy.orm import relationship, sessionmaker
from sqlalchemy.ext.declarative import declarative_base

from .capability_interfaces import CapabilityIF, CapabilityRequestIF, CapabilityExecutionIF, \
    CapabilitySequenceIF, ParameterIF
from .helpers import CapabilityStepType, CapabilityStep, CapabilitySequence
from .product_interfaces import FutureProductIF


class AbstractFile:
    """
    Abstract file is exactly that, an abstract concept of what a file is, to be
    returned from various non-filesystem places.
    """
    def __init__(self, filename: str, content: bytes):
        self.filename, self.content = filename, content

    def write_to(self, directory: Path):
        (directory / self.filename).write_bytes(self.content)

    def __json__(self, request):
        return {'filename': self.filename, 'content': self.content}

    @classmethod
    def from_path(cls, path: Path) -> 'AbstractFile':
        return cls(path.name, path.read_bytes())


class AbstractRequest:
    """
    Abstract File is an abstract concept that is used to create requests
    """
    workflow_name: str
    argument: json

    def __init__(self, workflow_name: str, argument: json):
        self.workflow_name, self.argument = workflow_name, argument

    def __json__(self, request):
        return {'workflow_name': self.workflow_name, 'argument': self.argument}


class WorkflowEventType(enum.Enum):
    """
    Kinds of events that a workflow can send. These four are derived from
    the many kinds that HTCondor can send.
    """
    SUBMITTED = 0
    EXECUTING = 1
    TERMINATED = 5
    OTHER = -1


class WorkflowEvent:
    """
    An event from the underlying workflow execution system, which is likely
    to be HTCondor.
    """

    def __init__(
            self,
            job_name: str,
            job_id: str,
            event_type: WorkflowEventType,
            timestamp: str,
            log: str,
            retval: int = None
    ):
        # NOTE: when you modify the properties here, also update the schema
        # in the .json package.
        self.job_name = job_name
        self.job_id = str(job_id)
        self.type = event_type
        self.timestamp = timestamp
        self.log = log
        self.retval = retval

    def json(self) -> str:
        d = {
            'job_name': self.job_name,
            'type': str(self.type),
            'timestamp': self.timestamp,
            'return_value': self.retval,
            'log': self.log
        }
        return json.dumps(d)

    def __str__(self):
        return f'WorkflowEvent with data {self.json()}'

    def __repr__(self):
        return f'<WorkflowEvent {self.__dict__}>'

    def __eq__(self, other):
        return self.job_name == other.job_name and \
               self.job_id == other.job_id and \
               self.type == other.type and \
               self.timestamp == other.timestamp and \
               self.log == other.log and \
               self.retval == other.retval


class CapabilityEvent:
    """
    An event from the underlying workflow execution system.
    """

    def __init__(self):
        raise NotImplementedError

    def json(self) -> str:
        raise NotImplementedError

    def __str__(self):
        return f'<CapabilityEvent with data{self.json()}>'

    def __repr__(self):
        return f'<CapabilityEvent {self.__dict__}>'

    def __eq__(self, other):
        raise NotImplementedError


Base = declarative_base(metaclass=abc.ABCMeta)


class Capability(Base, CapabilityIF):
    """
    A capability, which is a particular workflow setup, intended to accept
    a certain kind of product and some parameters and produce another product
    """
    __tablename__ = 'capabilities'
    id = sa.Column('capability_id', sa.Integer, primary_key=True)
    name = sa.Column('capability_name', sa.String)
    steps = sa.Column('capability_steps', sa.String)
    max_jobs = sa.Column('max_jobs', sa.Integer)

    @classmethod
    def from_file(cls, filename):
        self = cls()
        self.name, self.max_jobs, self.steps = cls.parse_capability_file(filename)
        return self

    @staticmethod
    def parse_capability_file(filename: str) -> Tuple[str, int, CapabilitySequenceIF]:
        with open(filename, 'r') as f:
            name, max_jobs = f.readline().split(' ')
            steps = []

            for line in f.readlines():
                step_type, step_value = line.strip().split(' ')
                steps.append(CapabilityStep(CapabilityStepType.from_string(step_type), step_value))

            return name, int(max_jobs), CapabilitySequence(steps)

    def create_request(
            self, parameters: ParameterIF, future_products: FutureProductIF, versions: str
    ):
        """
        Create a new request for this capability with specific options
        FIXME: Future products needs a string representation
        """
        return CapabilityRequest(
            capability=self.id,
            parameters=parameters.json(),
            future_products=str(future_products),
            versions=versions
        )

    def __str__(self):
        return (
            f'Capability object with name {self.name} and max simultaneous jobs of {self.max_jobs}'
            f'\nSequence: {self.steps}'
        )


class CapabilityRequest(Base, CapabilityRequestIF):
    """
    A capability request, which couples a capability to a product, representing
    the expectation of a new product given a set of parameters
    """
    __tablename__ = 'capability_requests'
    id = sa.Column('request_id', sa.Integer, primary_key=True)
    state = sa.Column('state', sa.String)
    capability = sa.Column(
        'capability',
        sa.Integer,
        sa.ForeignKey('capabilities.id')
    )
    parameters = sa.Column('parameters', sa.String)
    # FIXME: This needs to be changed to properly keep track of product locators.
    future_products = sa.Column('future_products', sa.String)
    versions = sa.Column('versions', sa.String)


class CapabilityExecution(Base, CapabilityExecutionIF):
    """
    Schema representation of a capability request's execution record
    """
    __tablename__ = 'capability_executions'
    id = sa.Column('execution_id', sa.Integer, primary_key=True)
    state = sa.Column('state', sa.String)
    capability_request = sa.Column(
        'capability_request_id',
        sa.Integer,
        sa.ForeignKey('capability_requests.capability_request_id')
    )
    current_step = sa.Column('current_step', sa.Integer)


class Workflow(Base):
    """
    A Workflow is a suite of tasks defined by templates that must be
    executed together with some user-submitted files.
    """
    __tablename__ = 'workflows'
    workflow_name = sa.Column('workflow_name', sa.String, primary_key=True)
    templates = relationship('WorkflowTemplate', backref='workflows')
    requests = relationship('WorkflowRequest', backref='workflows')

    def __json__(self, request):
        return {'workflow_name': self.workflow_name, 'templates': self.templates}

    def __repr__(self):
        return f'<Workflow workflow_name={self.workflow_name}>'

    def render_templates(self, argument: Dict, files: List[Path]) -> List[AbstractFile]:
        """
        Render the templates associated with this workflow
        :param argument: the workflow argument JSON
        :param files:    the files to be processed
        :return:         a list of rendered templates
        """

        # for file in files:
            # render file
        raise NotImplementedError


class WorkflowTemplate(Base):
    """
    Workflow Templates are Mustache-formatted files associated with a given Workflow
    """
    __tablename__ = 'workflow_templates'
    filename = sa.Column('filename', sa.String, primary_key=True)
    content = sa.Column('content', sa.LargeBinary, nullable=False)
    workflow_name = sa.Column('workflow_name', sa.String, sa.ForeignKey('workflows.workflow_name'), primary_key=True)

    def render(self, argument: Dict) -> AbstractFile:
        raise NotImplementedError('Add mustache to render this template')

    @property
    def template_file(self) -> AbstractFile:
        return AbstractFile(self.filename, self.content)

    @template_file.setter
    def template_file(self, file: AbstractFile):
        self.filename, self.content = file.filename, file.content

    def __json__(self, request):
        # Defer to the actual template file's contents for JSON conversion
        return self.template_file.__json__(request)

    def __repr__(self):
        return f'<WorkflowTemplate filename={self.filename} for workflow={self.workflow_name}>'


class WorkflowRequest(Base):
    """
    Workflow Requests invoke a particular workflow with a particular argument and some user-supplied files.
    """
    __tablename__ = 'workflow_requests'
    workflow_request_id = sa.Column('workflow_request_id', sa.Integer, primary_key=True, autoincrement=True)
    workflow_name = sa.Column('workflow_name', sa.String, sa.ForeignKey('workflows.workflow_name'))
    argument = sa.Column('argument', sa.JSON)
    files = relationship('WorkflowRequestFile', backref='request')

    @property
    def request(self) -> AbstractRequest:
        return AbstractRequest(self.workflow_name, self.argument)

    @request.setter
    def create_request(self, request: AbstractRequest):
        self.workflow_name, self.argument = request.workflow_name, request.argument

    def update_status(self, status: str):
        self.status = status

    def set_start_time(self, time: str):
        self.start_time = time

    def set_end_time(self, time:str):
        self.end_time = time

    def __repr__(self):
        return f'<WorkflowRequest workflow_request_id= {self.workflow_request_id}>'


class WorkflowRequestFile(Base):
    """
    A Workflow Request File is a file supplied by the user and attached to the request they have submitted.
    """
    __tablename__ = 'workflow_request_files'
    workflow_request_id = sa.Column('workflow_request_id', sa.Integer,
                                    sa.ForeignKey('workflow_requests.workflow_request_id'),
                                    primary_key=True)
    filename = sa.Column('filename', sa.String, primary_key=True)
    content = sa.Column('content', sa.LargeBinary, nullable=False)

    @property
    def file(self) -> AbstractFile:
        return AbstractFile(self.filename, self.content)

    @file.setter
    def file(self, file: AbstractFile):
        self.filename, self.content = file.filename, file.content

    def __repr__(self):
        return f'<WorkflowRequestFile filename={self.filename}>'


def get_engine():
    """
    Generate the SQL Alchemy engine for us, using Capo.
    :return:
    """
    capo = CapoConfig().settings('metadataDatabase')
    url = capo.jdbcUrl.replace('jdbc:', '').replace('://', f'://{capo.jdbcUsername}:{capo.jdbcPassword}@')
    return create_engine(url)


def get_session_factory(engine):
    """
    Generate the SQL Alchemy session factory for us, using the supplied engine
    :param engine:
    :return:
    """
    factory = sessionmaker()
    factory.configure(bind=engine)
    return factory
